from enum import Enum


class CommandId(str, Enum):
    OPEN = "myapp.open"
    CLOSE = "myapp.close"
    SAVE = "myapp.save"
    QUIT = "myapp.quit"

    UNDO = "myapp.undo"
    REDO = "myapp.redo"
    COPY = "myapp.copy"
    PASTE = "myapp.paste"
    CUT = "myapp.cut"

    def __str__(self) -> str:
        return self.value


class MenuId(str, Enum):
    FILE = "myapp/file"
    EDIT = "myapp/edit"

    def __str__(self) -> str:
        return self.value



from __future__ import annotations

import contextlib
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Literal,
    NamedTuple,
    TypeVar,
    overload,
)

from ._expressions import Name

if TYPE_CHECKING:
    import builtins
    from collections.abc import MutableMapping

T = TypeVar("T")
A = TypeVar("A")


class __missing:
    """Sentinel... done this way for the purpose of typing."""

    def __repr__(self) -> str:
        return "MISSING"


MISSING = __missing()


class ContextKeyInfo(NamedTuple):
    """Just a recordkeeping tuple.

    Retrieve all declared ContextKeys with ContextKeyInfo.info().
    """

    key: str
    type: type | None
    description: str | None
    namespace: builtins.type[ContextNamespace] | None


class ContextKey(Name, Generic[A, T]):
    """Context key name, default, description, and getter.

    This is intended to be used as class attribute in a `ContextNamespace`.
    This is a subclass of `Name`, and is therefore usable in an `Expression`.
    (see examples.)

    Parameters
    ----------
    default_value : Any, optional
        The default value for this key, by default MISSING
    description : str, optional
        Description of this key.  Useful for documentation, by default None
    getter : callable, optional
        Callable that receives an object and retrieves the current value for
        this key, by default None.
        For example, if this ContextKey represented the length of some list,
        (like the layerlist) it might look like
        `length = ContextKey(0, 'length of the list', lambda x: len(x))`
    id : str, optional
        Explicitly provide the `Name` string used when evaluating a context,
        by default the key will be taken as the attribute name to which this
        object is assigned as a class attribute:

    Examples
    --------
    >>> class MyNames(ContextNamespace):
    ...     some_key = ContextKey(0, "some description", lambda x: sum(x))

    >>> expr = MyNames.some_key > 5  # create an expression using this key

    these expressions can be later evaluated with some concrete context.

    >>> expr.eval({"some_key": 3})  # False
    >>> expr.eval({"some_key": 6})  # True
    """

    # This will catalog all ContextKeys that get instantiated, which provides
    # an easy way to organize documentation.
    # ContextKey.info() returns a list with info for all ContextKeys
    _info: ClassVar[list[ContextKeyInfo]] = []
    MISSING = MISSING

    def __init__(
        self,
        default_value: T | __missing = MISSING,
        description: str | None = None,
        getter: Callable[[A], T] | None = None,
        *,
        id: str = "",  # optional because of __set_name__
    ) -> None:
        bound = type(default_value) if default_value is not MISSING else None
        super().__init__(id or "", bound=bound)
        self._default_value = default_value
        self._getter = getter
        self._description = description
        self._owner: type[ContextNamespace] | None = None
        self._type = (
            type(default_value) if default_value not in (None, MISSING) else None
        )
        if id:
            self._store()

    def __str__(self) -> str:
        return self.id

    @classmethod
    def info(cls) -> list[ContextKeyInfo]:
        """Return list of all stored context keys."""
        return list(cls._info)

    def _store(self) -> None:
        self._info.append(
            ContextKeyInfo(self.id, self._type, self._description, self._owner)
        )

    def __set_name__(self, owner: type[ContextNamespace[A]], name: str) -> None:
        """Set the name for this key.

        (this happens when you instantiate this class as a class attribute).
        """
        if self.id:
            raise RuntimeError(
                f"Cannot change id of ContextKey (already {self.id!r})",
            )
        self._owner = owner
        self.id = name
        # recompile the code with the new name
        self._recompile()
        self._store()

    @overload
    def __get__(self, obj: Literal[None], objtype: type) -> ContextKey[A, T]:
        # When we __get__ from the class, we return ourself
        ...

    @overload
    def __get__(self, obj: ContextNamespace[A], objtype: type) -> T:
        # When we got from the object, we return the current value
        ...

    def __get__(
        self, obj: ContextNamespace[A] | None, objtype: type
    ) -> T | ContextKey[A, T] | None:
        """Get current value of the key in the associated context."""
        return self if obj is None else obj._context.get(self.id, MISSING)

    def __set__(self, obj: ContextNamespace[A], value: T) -> None:
        """Set current value of the key in the associated context."""
        obj._context[self.id] = value

    def __delete__(self, obj: ContextNamespace[A]) -> None:
        """Delete key from the associated context."""
        del obj._context[self.id]


class ContextNamespaceMeta(type):
    """Metaclass that finds all ContextNamespace members."""

    _members_map_: dict[str, ContextKey]

    def __new__(cls, clsname: str, bases: tuple, attrs: dict) -> ContextNamespaceMeta:
        """Create a new ContextNamespace class."""
        new_cls = super().__new__(cls, clsname, bases, attrs)
        new_cls._members_map_ = {
            k: v for k, v in attrs.items() if isinstance(v, ContextKey)
        }
        return new_cls

    @property
    def __members__(self) -> MappingProxyType[str, ContextKey]:
        return MappingProxyType(self._members_map_)

    def __dir__(self) -> list[str]:  # pragma: no cover
        return [
            "__class__",
            "__doc__",
            "__members__",
            "__module__",
            *list(self._members_map_),
        ]


class ContextNamespace(Generic[A], metaclass=ContextNamespaceMeta):
    """A collection of related keys in a context.

    meant to be subclassed, with `ContextKeys` as class attributes.
    """

    def __init__(self, context: MutableMapping) -> None:
        self._context = context

        # on instantiation we create an index of defaults and value-getters
        # to speed up retrieval later
        self._defaults: dict[str, Any] = {}  # default values per key
        self._getters: dict[str, Callable[[A], Any]] = {}  # value getters
        for name, ctxkey in type(self).__members__.items():
            self._defaults[name] = ctxkey._default_value
            if ctxkey._default_value is not MISSING:
                context[ctxkey.id] = ctxkey._default_value
            if callable(ctxkey._getter):
                self._getters[name] = ctxkey._getter

    def reset(self, key: str) -> None:
        """Reset keys to its default."""
        val = self._defaults[key]
        if val is MISSING:
            with contextlib.suppress(KeyError):
                delattr(self, key)
        else:
            setattr(self, key, self._defaults[key])

    def reset_all(self) -> None:
        """Reset all keys to their defaults."""
        for key in self._defaults:
            self.reset(key)

    def dict(self) -> dict:
        """Return all keys in this namespace."""
        return {k: getattr(self, k) for k in type(self).__members__}

    def __repr__(self) -> str:
        import pprint

        return pprint.pformat(self.dict())



"""App-model registries, such as menus, keybindings, commands."""

from ._commands_reg import CommandsRegistry, RegisteredCommand
from ._keybindings_reg import KeyBindingsRegistry
from ._menus_reg import MenusRegistry
from ._register import register_action

__all__ = [
    "CommandsRegistry",
    "KeyBindingsRegistry",
    "MenusRegistry",
    "RegisteredCommand",
    "register_action",
]



import re
from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel, Field, model_validator

from app_model.types._constants import OperatingSystem

from ._key_codes import KeyChord, KeyCode, KeyMod

if TYPE_CHECKING:
    from pydantic.annotated_handlers import GetCoreSchemaHandler
    from pydantic_core import core_schema


class SimpleKeyBinding(BaseModel):
    """Represent a simple combination modifier(s) and a key, e.g. Ctrl+A."""

    ctrl: bool = Field(False, description='Whether the "Ctrl" modifier is active.')
    shift: bool = Field(False, description='Whether the "Shift" modifier is active.')
    alt: bool = Field(False, description='Whether the "Alt" modifier is active.')
    meta: bool = Field(False, description='Whether the "Meta" modifier is active.')
    key: Optional[KeyCode] = Field(
        None, description="The key that is pressed (e.g. `KeyCode.A`)"
    )

    # def hash_code(self) -> str:
    # used by vscode for caching during keybinding resolution

    def is_modifier_key(self) -> bool:
        """Return true if this is a modifier key."""
        return self.key in (
            KeyCode.Alt,
            KeyCode.Shift,
            KeyCode.Ctrl,
            KeyCode.Meta,
            KeyCode.UNKNOWN,
        )

    def __str__(self) -> str:
        """Get a normalized string representation (constant to all OSes) of this SimpleKeyBinding."""
        out = ""
        if self.ctrl:
            out += "Ctrl+"
        if self.shift:
            out += "Shift+"
        if self.alt:
            out += "Alt+"
        if self.meta:
            out += "Meta+"
        if self.key:
            out += str(self.key)
        return out

    def __eq__(self, other: Any) -> bool:
        # sourcery skip: remove-unnecessary-cast
        if not isinstance(other, SimpleKeyBinding):
            try:
                if (other := SimpleKeyBinding._parse_input(other)) is None:
                    return NotImplemented
            except TypeError:  # pragma: no cover  # can happen with pydantic v2
                return NotImplemented
        return bool(
            self.ctrl == other.ctrl
            and self.shift == other.shift
            and self.alt == other.alt
            and self.meta == other.meta
            and self.key == other.key
        )

    @classmethod
    def from_str(cls, key_str: str) -> "SimpleKeyBinding":
        """Parse a string into a SimpleKeyBinding."""
        mods, remainder = _parse_modifiers(key_str.strip())
        key = KeyCode.from_string(remainder)
        return cls(**mods, key=key)

    @classmethod
    def from_int(
        cls, key_int: int, os: Optional[OperatingSystem] = None
    ) -> "SimpleKeyBinding":
        """Create a SimpleKeyBinding from an integer."""
        ctrl_cmd = bool(key_int & KeyMod.CtrlCmd)
        win_ctrl = bool(key_int & KeyMod.WinCtrl)
        shift = bool(key_int & KeyMod.Shift)
        alt = bool(key_int & KeyMod.Alt)

        os = OperatingSystem.current() if os is None else os
        ctrl = win_ctrl if os.is_mac else ctrl_cmd
        meta = ctrl_cmd if os.is_mac else win_ctrl
        key = key_int & 0x000000FF  # keycode mask

        return cls(ctrl=ctrl, shift=shift, alt=alt, meta=meta, key=key)

    def __int__(self) -> int:
        return int(self.to_int())

    def __hash__(self) -> int:
        return hash((self.ctrl, self.shift, self.alt, self.meta, self.key))

    def to_int(self, os: Optional[OperatingSystem] = None) -> int:
        """Convert this SimpleKeyBinding to an integer representation."""
        os = OperatingSystem.current() if os is None else os
        mods: KeyMod = KeyMod.NONE
        if self.ctrl:
            mods |= KeyMod.WinCtrl if os.is_mac else KeyMod.CtrlCmd
        if self.shift:
            mods |= KeyMod.Shift
        if self.alt:
            mods |= KeyMod.Alt
        if self.meta:
            mods |= KeyMod.CtrlCmd if os.is_mac else KeyMod.WinCtrl
        return mods | (self.key or 0)

    def _mods2keycodes(self) -> list[KeyCode]:
        """Create KeyCode instances list of modifiers from this SimpleKeyBinding."""
        mods = []
        if self.ctrl:
            mods.append(KeyCode.Ctrl)
        if self.shift:
            mods.append(KeyCode.Shift)
        if self.alt:
            mods.append(KeyCode.Alt)
        if self.meta:
            mods.append(KeyCode.Meta)
        return mods

    def to_text(
        self,
        os: Optional[OperatingSystem] = None,
        use_symbols: bool = False,
        joinchar: str = "+",
    ) -> str:
        """Get a user-facing string representation of this SimpleKeyBinding.

        Optionally, the string representation can be constructed with symbols
        like ↵ for Enter or OS specific ones like ⌘ for Meta on MacOS. If no symbols
        should be used, the string representation will use the OS specific names
        for the keys like `Cmd` for Meta or `Option` for Ctrl on MacOS.

        Also, a join character can be defined. By default `+` is used.
        """
        os = OperatingSystem.current() if os is None else os
        keybinding_elements = [*self._mods2keycodes()]
        if self.key:
            keybinding_elements.append(self.key)

        return joinchar.join(
            kbe.os_symbol(os=os) if use_symbols else kbe.os_name(os=os)
            for kbe in keybinding_elements
        )

    @classmethod
    def _parse_input(cls, v: Any) -> "SimpleKeyBinding":
        if isinstance(v, SimpleKeyBinding):
            return v
        if isinstance(v, str):
            return cls.from_str(v)
        if isinstance(v, int):
            return cls.from_int(v)
        raise TypeError(f"invalid type: {type(v)}")

    @model_validator(mode="before")
    @classmethod
    def _model_val(cls, val: "SimpleKeyBinding") -> "SimpleKeyBinding":
        if not isinstance(val, (SimpleKeyBinding, dict)):
            return cls._parse_input(val)
        return val


MIN1 = {"min_length": 1}


class KeyBinding:
    """KeyBinding.  May be a multi-part "Chord" (e.g. 'Ctrl+K Ctrl+C').

    This is the primary representation of a fully resolved keybinding. For consistency
    in the downstream API, it should be preferred to
    [`SimpleKeyBinding`][app_model.types.SimpleKeyBinding], even
    when there is only a single part in the keybinding (i.e. when it is not a chord.)

    Chords (two separate keypress actions) are expressed as a string by separating
    the two keypress codes with a space. For example, 'Ctrl+K Ctrl+C'.

    Parameters
    ----------
    parts : List[SimpleKeyBinding]
        The parts of the keybinding.  There must be at least one part.
    """

    parts: list[SimpleKeyBinding] = Field(..., **MIN1)  # type: ignore

    def __init__(self, *, parts: list[SimpleKeyBinding]) -> None:
        self.parts = parts

    def __str__(self) -> str:
        """Get a normalized string representation (constant to all OSes) of this KeyBinding."""
        return " ".join(str(part) for part in self.parts)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} at {hex(id(self))}: {self}>"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, KeyBinding):
            return self.parts == other.parts
        return NotImplemented

    def __len__(self) -> int:
        return len(self.parts)

    @property
    def part0(self) -> SimpleKeyBinding:
        """Return the first part of this keybinding.

        All keybindings have at least one part.
        """
        return self.parts[0]

    @classmethod
    def from_str(cls, key_str: str) -> "KeyBinding":
        """Parse a string into a SimpleKeyBinding."""
        parts = [SimpleKeyBinding.from_str(part) for part in key_str.split()]
        return cls(parts=parts)

    @classmethod
    def from_int(
        cls, key_int: int, os: Optional[OperatingSystem] = None
    ) -> "KeyBinding":
        """Create a KeyBinding from an integer."""
        # a multi keybinding is represented as an integer
        # with the first_part in the lowest 16 bits,
        # the second_part in the next 16 bits, etc.
        first_part = key_int & 0x0000FFFF
        chord_part = (key_int & 0xFFFF0000) >> 16
        if chord_part != 0:
            return cls(
                parts=[
                    SimpleKeyBinding.from_int(first_part, os),
                    SimpleKeyBinding.from_int(chord_part, os),
                ]
            )
        return cls(parts=[SimpleKeyBinding.from_int(first_part, os)])

    def to_int(self, os: Optional[OperatingSystem] = None) -> int:
        """Convert this KeyBinding to an integer representation."""
        if len(self.parts) > 2:  # pragma: no cover
            raise NotImplementedError(
                "Cannot represent chords with more than 2 parts as int"
            )
        os = OperatingSystem.current() if os is None else os
        parts = [part.to_int(os) for part in self.parts]
        if len(parts) == 2:
            return KeyChord(*parts)
        return parts[0]

    def to_text(
        self,
        os: Optional[OperatingSystem] = None,
        use_symbols: bool = False,
        joinchar: str = "+",
    ) -> str:
        """Get a text representation of this KeyBinding.

        Optionally, the string representation can be constructed with symbols
        like ↵ for Enter or OS specific ones like ⌘ for Meta on MacOS. If no symbols
        should be used, the string representation will use the OS specific names
        for the keys like `Cmd` for Meta or `Option` for Ctrl on MacOS.

        Also, a join character can be defined. By default `+` is used.
        """
        return " ".join(
            part.to_text(os=os, use_symbols=use_symbols, joinchar=joinchar)
            for part in self.parts
        )

    def __int__(self) -> int:
        return int(self.to_int())

    def __hash__(self) -> int:
        return hash(tuple(self.parts))

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type, handler: "GetCoreSchemaHandler"
    ) -> "core_schema.CoreSchema":
        from pydantic_core import core_schema

        return core_schema.no_info_plain_validator_function(
            cls.validate, serialization=core_schema.to_string_ser_schema()
        )

    @classmethod
    def validate(cls, v: Any) -> "KeyBinding":
        """Validate a SimpleKeyBinding."""
        if isinstance(v, KeyBinding):
            return v
        if isinstance(v, SimpleKeyBinding):
            return cls(parts=[v])
        if isinstance(v, int):
            return cls.from_int(v)
        if isinstance(v, str):
            return cls.from_str(v)
        raise TypeError("invalid keybinding")  # pragma: no cover


_re_ctrl = re.compile(r"(ctrl|control|ctl|⌃|\^)[\+|\-]")
_re_shift = re.compile(r"(shift|⇧)[\+|\-]")
_re_alt = re.compile(r"(alt|opt|option|⌥)[\+|\-]")
_re_meta = re.compile(r"(meta|super|win|windows|⊞|cmd|command|⌘)[\+|\-]")


def _parse_modifiers(input: str) -> tuple[dict[str, bool], str]:
    """Parse modifiers from a string (case insensitive).

    modifiers must start at the beginning of the string, and be separated by
    "+" or "-".  e.g. "ctrl+shift+alt+K" or "Ctrl-Cmd-K"
    """
    remainder = input.lower()

    patterns = {"ctrl": _re_ctrl, "shift": _re_shift, "alt": _re_alt, "meta": _re_meta}
    mods = dict.fromkeys(patterns, False)
    while True:
        saw_modifier = False
        for key, ptrn in patterns.items():
            if m := ptrn.match(remainder):
                remainder = remainder[m.span()[1] :]
                mods[key] = True
                saw_modifier = True
                break
        if not saw_modifier:
            break

    return mods, remainder



# mypy: disable-error-code="operator"
# pyright: reportOperatorIssue=false
from unittest.mock import patch

from qtpy.QtCore import Qt
from qtpy.QtGui import QKeySequence

from app_model.backends.qt import (
    _qkeymap,
    qkey2modelkey,
    qkeysequence2modelkeybinding,
    qmods2modelmods,
)
from app_model.backends.qt._qkeymap import modelkey2qkey
from app_model.types import KeyBinding, KeyCode, KeyCombo, KeyMod

# stuff we don't know how to deal with yet


def test_modelkey_lookup() -> None:
    assert modelkey2qkey(KeyCode.KeyM) == Qt.Key.Key_M

    with patch.object(_qkeymap, "MAC", True):
        with patch.object(_qkeymap, "_mac_ctrl_meta_swapped", return_value=False):
            assert modelkey2qkey(KeyCode.Ctrl) == Qt.Key.Key_Control
            assert modelkey2qkey(KeyCode.Meta) == Qt.Key.Key_Meta
        with patch.object(_qkeymap, "_mac_ctrl_meta_swapped", return_value=True):
            assert modelkey2qkey(KeyCode.Meta) == Qt.Key.Key_Control
            assert modelkey2qkey(KeyCode.Ctrl) == Qt.Key.Key_Meta

    with patch.object(_qkeymap, "MAC", False):
        assert modelkey2qkey(KeyCode.Ctrl) == Qt.Key.Key_Control
        assert modelkey2qkey(KeyCode.Meta) == Qt.Key.Key_Meta


def test_qkey_lookup() -> None:
    for keyname in (k for k in dir(Qt.Key) if k.startswith("Key")):
        key = getattr(Qt.Key, keyname)
        assert isinstance(qkey2modelkey(key), (KeyCode, KeyCombo))

    assert qkey2modelkey(Qt.Key.Key_M) == KeyCode.KeyM

    with patch.object(_qkeymap, "MAC", True):
        with patch.object(_qkeymap, "_mac_ctrl_meta_swapped", return_value=False):
            assert qkey2modelkey(Qt.Key.Key_Control) == KeyCode.Ctrl
            assert qkey2modelkey(Qt.Key.Key_Meta) == KeyCode.Meta
        with patch.object(_qkeymap, "_mac_ctrl_meta_swapped", return_value=True):
            assert qkey2modelkey(Qt.Key.Key_Control) == KeyCode.Meta
            assert qkey2modelkey(Qt.Key.Key_Meta) == KeyCode.Ctrl

    with patch.object(_qkeymap, "MAC", False):
        assert qkey2modelkey(Qt.Key.Key_Control) == KeyCode.Ctrl
        assert qkey2modelkey(Qt.Key.Key_Meta) == KeyCode.Meta


def test_qmod_lookup() -> None:
    assert qmods2modelmods(Qt.KeyboardModifier.ShiftModifier) == KeyMod.Shift
    assert qmods2modelmods(Qt.KeyboardModifier.AltModifier) == KeyMod.Alt

    with patch.object(_qkeymap, "MAC", True):
        with patch.object(_qkeymap, "_mac_ctrl_meta_swapped", return_value=False):
            assert (
                qmods2modelmods(Qt.KeyboardModifier.ControlModifier) == KeyMod.WinCtrl
            )
            assert qmods2modelmods(Qt.KeyboardModifier.MetaModifier) == KeyMod.CtrlCmd

        with patch.object(_qkeymap, "_mac_ctrl_meta_swapped", return_value=True):
            assert (
                qmods2modelmods(Qt.KeyboardModifier.ControlModifier) == KeyMod.CtrlCmd
            )
            assert qmods2modelmods(Qt.KeyboardModifier.MetaModifier) == KeyMod.WinCtrl

    with patch.object(_qkeymap, "MAC", False):
        assert qmods2modelmods(Qt.KeyboardModifier.ControlModifier) == KeyMod.CtrlCmd
        assert qmods2modelmods(Qt.KeyboardModifier.MetaModifier) == KeyMod.WinCtrl


def test_qkeysequence2modelkeybinding() -> None:
    seq = QKeySequence(
        Qt.Modifier.SHIFT | Qt.Key.Key_M, Qt.KeyboardModifier.NoModifier | Qt.Key.Key_K
    )
    app_key = KeyBinding(parts=[KeyMod.Shift | KeyCode.KeyM, KeyCode.KeyK])
    assert qkeysequence2modelkeybinding(seq) == app_key

    seq = QKeySequence(
        Qt.Modifier.ALT | Qt.Key.Key_M, Qt.KeyboardModifier.NoModifier | Qt.Key.Key_K
    )
    app_key = KeyBinding(parts=[KeyMod.Alt | KeyCode.KeyM, KeyCode.KeyK])
    assert qkeysequence2modelkeybinding(seq) == app_key

    with patch.object(_qkeymap, "MAC", True):
        with patch.object(_qkeymap, "_mac_ctrl_meta_swapped", return_value=False):
            # on Macs, unswapped, Meta -> Cmd
            seq = QKeySequence(
                Qt.Modifier.META | Qt.Key.Key_M,
                Qt.KeyboardModifier.NoModifier | Qt.Key.Key_K,
            )
            app_key = KeyBinding(parts=[KeyMod.CtrlCmd | KeyCode.KeyM, KeyCode.KeyK])
            assert qkeysequence2modelkeybinding(seq) == app_key

            # on Macs, unswapped, Ctrl -> Ctrl
            seq = QKeySequence(
                Qt.Modifier.CTRL | Qt.Key.Key_M,
                Qt.KeyboardModifier.NoModifier | Qt.Key.Key_K,
            )
            app_key = KeyBinding(parts=[KeyMod.WinCtrl | KeyCode.KeyM, KeyCode.KeyK])
            assert qkeysequence2modelkeybinding(seq) == app_key

            seq = QKeySequence(
                Qt.Modifier.META | Qt.Key.Key_Meta,
                Qt.Modifier.CTRL | Qt.Key.Key_Control,
            )
            app_key = KeyBinding(
                parts=[KeyMod.CtrlCmd | KeyCode.Meta, KeyMod.WinCtrl | KeyCode.Ctrl]
            )
            assert qkeysequence2modelkeybinding(seq) == app_key

        with patch.object(_qkeymap, "_mac_ctrl_meta_swapped", return_value=True):
            # on Mac swapped, Ctrl -> Meta/Cmd
            seq = QKeySequence(
                Qt.Modifier.CTRL | Qt.Key.Key_M,
                Qt.KeyboardModifier.NoModifier | Qt.Key.Key_K,
            )
            app_key = KeyBinding(parts=[KeyMod.CtrlCmd | KeyCode.KeyM, KeyCode.KeyK])
            assert qkeysequence2modelkeybinding(seq) == app_key

            # on Mac swapped, Meta/Cmd -> Ctrl
            seq = QKeySequence(
                Qt.Modifier.META | Qt.Key.Key_M,
                Qt.KeyboardModifier.NoModifier | Qt.Key.Key_K,
            )
            app_key = KeyBinding(parts=[KeyMod.WinCtrl | KeyCode.KeyM, KeyCode.KeyK])
            assert qkeysequence2modelkeybinding(seq) == app_key

            seq = QKeySequence(
                Qt.Modifier.META | Qt.Key.Key_Meta,
                Qt.Modifier.CTRL | Qt.Key.Key_Control,
            )
            app_key = KeyBinding(
                parts=[KeyMod.WinCtrl | KeyCode.Ctrl, KeyMod.CtrlCmd | KeyCode.Meta]
            )
            assert qkeysequence2modelkeybinding(seq) == app_key

    with patch.object(_qkeymap, "MAC", False):
        # on Win/Unix, Ctrl -> Ctrl
        seq = QKeySequence(
            Qt.Modifier.CTRL | Qt.Key.Key_M,
            Qt.KeyboardModifier.NoModifier | Qt.Key.Key_K,
        )
        app_key = KeyBinding(parts=[KeyMod.CtrlCmd | KeyCode.KeyM, KeyCode.KeyK])
        assert qkeysequence2modelkeybinding(seq) == app_key

        # on Win, Meta -> Win, on Unix, Meta -> Super
        seq = QKeySequence(
            Qt.Modifier.META | Qt.Key.Key_M,
            Qt.KeyboardModifier.NoModifier | Qt.Key.Key_K,
        )
        app_key = KeyBinding(parts=[KeyMod.WinCtrl | KeyCode.KeyM, KeyCode.KeyK])
        assert qkeysequence2modelkeybinding(seq) == app_key

        seq = QKeySequence(
            Qt.Modifier.META | Qt.Key.Key_Meta,
            Qt.Modifier.CTRL | Qt.Key.Key_Control,
        )
        app_key = KeyBinding(
            parts=[KeyMod.WinCtrl | KeyCode.Meta, KeyMod.CtrlCmd | KeyCode.Ctrl]
        )
        assert qkeysequence2modelkeybinding(seq) == app_key



"""Abstraction on expressions, and contexts in which to evaluate them."""

from ._context import Context, app_model_context, create_context, get_context
from ._context_keys import ContextKey, ContextKeyInfo, ContextNamespace
from ._expressions import (
    BinOp,
    BoolOp,
    Compare,
    Constant,
    Expr,
    IfExp,
    Name,
    UnaryOp,
    parse_expression,
    safe_eval,
)

__all__ = [
    "BinOp",
    "BoolOp",
    "Compare",
    "Constant",
    "Context",
    "ContextKey",
    "ContextKeyInfo",
    "ContextNamespace",
    "Expr",
    "IfExp",
    "Name",
    "UnaryOp",
    "app_model_context",
    "create_context",
    "get_context",
    "parse_expression",
    "safe_eval",
]



from __future__ import annotations

import contextlib
import os
import sys
from collections.abc import Iterable, MutableMapping
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    Optional,
    overload,
)

import in_n_out as ino
from psygnal import Signal

from .expressions import Context, app_model_context
from .registries import (
    CommandsRegistry,
    KeyBindingsRegistry,
    MenusRegistry,
    register_action,
)
from .types import (
    Action,
)

if TYPE_CHECKING:
    from typing import Callable

    from .expressions import Expr
    from .registries._register import CommandDecorator
    from .types import (
        DisposeCallable,
        IconOrDict,
        KeyBindingRuleOrDict,
        MenuRuleOrDict,
    )


class Application:
    """Full application model.

    This is the top level object that comprises all of the registries, and other
    app-namespace specific objects.

    Parameters
    ----------
    name : str
        A name for this application.
    raise_synchronous_exceptions : bool
        Whether to raise exceptions that occur while executing commands synchronously,
        by default False. This is settable after instantiation, and can also be
        controlled per execution by calling `result.result()` on the future object
        returned from the `execute_command` method.
    commands_reg_class : Type[CommandsRegistry]
        (Optionally) override the class to use when creating the CommandsRegistry
    menus_reg_class : Type[MenusRegistry]
        (Optionally) override the class to use when creating the MenusRegistry
    keybindings_reg_class : Type[KeyBindingsRegistry]
        (Optionally) override the class to use when creating the KeyBindingsRegistry
    injection_store_class : Type[ino.Store]
        (Optionally) override the class to use when creating the injection Store
    context : Context | MutableMapping | None
        (Optionally) provide a context to use for this application. If a
        `MutableMapping` is provided, it will be used to create a `Context` instance.
        If `None` (the default), a new `Context` instance will be created.

    Attributes
    ----------
    commands : CommandsRegistry
        The Commands Registry for this application.
    menus : MenusRegistry
        The Menus Registry for this application.
    keybindings : KeyBindingsRegistry
        The KeyBindings Registry for this application.
    injection_store : in_n_out.Store
        The Injection Store for this application.
    context : Context
        The Context for this application.
    """

    destroyed = Signal(str)
    _instances: ClassVar[dict[str, Application]] = {}

    def __init__(
        self,
        name: str,
        *,
        raise_synchronous_exceptions: bool = False,
        commands_reg_class: type[CommandsRegistry] = CommandsRegistry,
        menus_reg_class: type[MenusRegistry] = MenusRegistry,
        keybindings_reg_class: type[KeyBindingsRegistry] = KeyBindingsRegistry,
        injection_store_class: type[ino.Store] = ino.Store,
        context: Context | MutableMapping | None = None,
    ) -> None:
        self._name = name
        if name in Application._instances:
            raise ValueError(
                f"Application {name!r} already exists. Retrieve it with "
                f"`Application.get_or_create({name!r})`."
            )
        Application._instances[name] = self

        if context is None:
            context = Context()
        elif isinstance(context, MutableMapping):
            context = Context(context)
        if not isinstance(context, Context):
            raise TypeError(
                f"context must be a Context or MutableMapping, got {type(context)}"
            )
        self._context = context
        self._context.update(app_model_context())

        self._context["is_linux"] = sys.platform.startswith("linux")
        self._context["is_mac"] = sys.platform == "darwin"
        self._context["is_windows"] = os.name == "nt"

        self._injection_store = injection_store_class.create(name)
        self._commands = commands_reg_class(
            self.injection_store,
            raise_synchronous_exceptions=raise_synchronous_exceptions,
        )
        self._menus = menus_reg_class()
        self._keybindings = keybindings_reg_class()

        self.injection_store.on_unannotated_required_args = "ignore"

        self._registered_actions: dict[str, Action] = {}
        self._disposers: list[tuple[str, DisposeCallable]] = []

    @property
    def raise_synchronous_exceptions(self) -> bool:
        """Whether to raise synchronous exceptions."""
        return self._commands._raise_synchronous_exceptions

    @raise_synchronous_exceptions.setter
    def raise_synchronous_exceptions(self, value: bool) -> None:
        self._commands._raise_synchronous_exceptions = value

    @property
    def commands(self) -> CommandsRegistry:
        """Return the [`CommandsRegistry`][app_model.registries.CommandsRegistry]."""
        return self._commands

    @property
    def menus(self) -> MenusRegistry:
        """Return the [`MenusRegistry`][app_model.registries.MenusRegistry]."""
        return self._menus

    @property
    def keybindings(self) -> KeyBindingsRegistry:
        """Return the [`KeyBindingsRegistry`][app_model.registries.KeyBindingsRegistry]."""  # noqa E501
        return self._keybindings

    @property
    def injection_store(self) -> ino.Store:
        """Return the `in_n_out.Store` instance associated with this `Application`."""
        return self._injection_store

    @property
    def context(self) -> Context:
        """Return the [`Context`][app_model.expressions.Context] for this application."""  # noqa E501
        return self._context

    @classmethod
    def get_or_create(cls, name: str) -> Application:
        """Get app named `name` or create and return a new one if it doesn't exist."""
        return cls._instances[name] if name in cls._instances else cls(name)

    @classmethod
    def get_app(cls, name: str) -> Optional[Application]:
        """Return app named `name` or None if it doesn't exist."""
        return cls._instances.get(name)

    @classmethod
    def destroy(cls, name: str) -> None:
        """Destroy the `Application` named `name`.

        This will call [`dispose()`][app_model.Application.dispose], destroy the
        injection store, and remove the application from the list of stored
        application names (allowing the name to be reused).
        """
        if name not in cls._instances:
            return  # pragma: no cover
        app = cls._instances.pop(name)
        app.dispose()
        app.injection_store.destroy(name)
        app.destroyed.emit(app.name)

    @property
    def name(self) -> str:
        """Return the name of this `Application`."""
        return self._name

    def __repr__(self) -> str:
        return f"Application({self.name!r})"

    def dispose(self) -> None:
        """Dispose this `Application`.

        This calls all disposers functions (clearing all registries).
        """
        while self._disposers:
            with contextlib.suppress(Exception):
                self._disposers.pop()[1]()

    @overload
    def register_action(self, action: Action) -> DisposeCallable: ...

    @overload
    def register_action(
        self,
        action: str,
        title: str,
        *,
        callback: Literal[None] = ...,
        category: str | None = ...,
        tooltip: str | None = ...,
        icon: IconOrDict | None = ...,
        enablement: Expr | None = ...,
        menus: list[MenuRuleOrDict] | None = ...,
        keybindings: list[KeyBindingRuleOrDict] | None = ...,
        palette: bool = True,
    ) -> CommandDecorator: ...

    @overload
    def register_action(
        self,
        action: str,
        title: str,
        *,
        callback: Callable[..., Any],
        category: str | None = ...,
        tooltip: str | None = ...,
        icon: IconOrDict | None = ...,
        enablement: Expr | None = ...,
        menus: list[MenuRuleOrDict] | None = ...,
        keybindings: list[KeyBindingRuleOrDict] | None = ...,
        palette: bool = True,
    ) -> DisposeCallable: ...

    def register_action(
        self,
        action: str | Action,
        title: str | None = None,
        *,
        callback: Callable[..., Any] | None = None,
        category: str | None = None,
        tooltip: str | None = None,
        icon: IconOrDict | None = None,
        enablement: Expr | None = None,
        menus: list[MenuRuleOrDict] | None = None,
        keybindings: list[KeyBindingRuleOrDict] | None = None,
        palette: bool = True,
    ) -> CommandDecorator | DisposeCallable:
        """Register [`Action`][app_model.Action] instance with this application.

        An [`Action`][app_model.Action] is the complete representation of a command,
        including information about where and whether it appears in menus and optional
        keybinding rules.

        See [`register_action`][app_model.register_action] for complete
        details on this function.
        """
        if isinstance(action, Action):
            return register_action(self, action)

        return register_action(
            self,
            id_or_action=action,
            title=title,  # type: ignore
            callback=callback,  # type: ignore
            category=category,
            tooltip=tooltip,
            icon=icon,
            enablement=enablement,
            menus=menus,
            keybindings=keybindings,
            palette=palette,
        )

    def register_actions(self, actions: Iterable[Action]) -> DisposeCallable:
        """Register multiple [`Action`][app_model.Action] instances with this app.

        Returns a function that may be called to undo the registration of `actions`.
        """
        d = [self.register_action(action) for action in actions]

        def _dispose() -> None:
            while d:
                d.pop()()

        return _dispose

    @property
    def registered_actions(self) -> MappingProxyType[str, Action]:
        """Return a Mapping of id->Action object for all registered actions.

        Note that this only includes actions that were registered using
        `register_action`.  Commands registered directly via
        `Application.commands.register_action` will not be included in this mapping.
        """
        return MappingProxyType(self._registered_actions)

    def _register_action_obj(self, action: Action) -> DisposeCallable:
        """Register an Action object. Return a function that unregisters the action.

        Helper for `register_action()`.
        """
        # register commands
        disposers = [self.commands.register_action(action)]
        # register keybindings
        if dk := self.keybindings.register_action_keybindings(action):
            disposers.append(dk)
        # register menus
        if dm := self.menus.append_action_menus(action):
            disposers.append(dm)

        # remember the action object as a whole.
        # note that commands.register_action will have raised an exception
        # if the action.id is already registered, so we can assume that
        # the keys are unique.
        self._registered_actions[action.id] = action

        # create a function that will dispose of all the disposers
        def _dispose() -> None:
            self._registered_actions.pop(action.id, None)
            for d in disposers:
                d()

        self._disposers.append((action.id, _dispose))
        return _dispose



# Copyright (C) 2013 Riverbank Computing Limited.
# Copyright (C) 2022 The Qt Company Ltd.
# SPDX-License-Identifier: LicenseRef-Qt-Commercial OR BSD-3-Clause
# pyright: reportOptionalMemberAccess=false

from fonticon_fa6 import FA6S
from qtpy.QtCore import QFile, QFileInfo, QSaveFile, Qt, QTextStream
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import (
    QAction,
    QApplication,
    QFileDialog,
    QMainWindow,
    QMessageBox,
    QTextEdit,
)
from superqt import fonticon


class MainWindow(QMainWindow):
    def __init__(self) -> None:
        super().__init__()

        self._cur_file = ""

        self._text_edit = QTextEdit()
        self.setCentralWidget(self._text_edit)

        self.create_actions()
        self.create_menus()
        self.create_tool_bars()
        self.create_status_bar()

        self.set_current_file("")

    def new_file(self) -> None:
        if self.maybe_save():
            self._text_edit.clear()
            self.set_current_file("")

    def open(self) -> None:
        if self.maybe_save():
            fileName, _filtr = QFileDialog.getOpenFileName(self)
            if fileName:
                self.load_file(fileName)

    def save(self) -> bool:
        return self.save_file(self._cur_file) if self._cur_file else self.save_as()

    def save_as(self) -> bool:
        fileName, _filtr = QFileDialog.getSaveFileName(self)
        if fileName:
            return self.save_file(fileName)

        return False

    def about(self) -> None:
        QMessageBox.about(
            self,
            "About Application",
            "The <b>Application</b> example demonstrates how to write "
            "modern GUI applications using Qt, with a menu bar, "
            "toolbars, and a status bar.",
        )

    def create_actions(self) -> None:
        self._new_act = QAction(
            fonticon.icon(FA6S.file_circle_plus),
            "&New",
            self,
            shortcut=QKeySequence.StandardKey.New,
            statusTip="Create a new file",
            triggered=self.new_file,
        )

        self._open_act = QAction(
            fonticon.icon(FA6S.folder_open),
            "&Open...",
            self,
            shortcut=QKeySequence.StandardKey.Open,
            statusTip="Open an existing file",
            triggered=self.open,
        )

        self._save_act = QAction(
            fonticon.icon(FA6S.floppy_disk),
            "&Save",
            self,
            shortcut=QKeySequence.StandardKey.Save,
            statusTip="Save the document to disk",
            triggered=self.save,
        )

        self._save_as_act = QAction(
            "Save &As...",
            self,
            shortcut=QKeySequence.StandardKey.SaveAs,
            statusTip="Save the document under a new name",
            triggered=self.save_as,
        )

        self._exit_act = QAction(
            "E&xit",
            self,
            shortcut="Ctrl+Q",
            statusTip="Exit the application",
            triggered=self.close,
        )

        self._cut_act = QAction(
            fonticon.icon(FA6S.scissors),
            "Cu&t",
            self,
            shortcut=QKeySequence.StandardKey.Cut,
            statusTip="Cut the current selection's contents to the clipboard",
            triggered=self._text_edit.cut,
        )

        self._copy_act = QAction(
            fonticon.icon(FA6S.copy),
            "&Copy",
            self,
            shortcut=QKeySequence.StandardKey.Copy,
            statusTip="Copy the current selection's contents to the clipboard",
            triggered=self._text_edit.copy,
        )

        self._paste_act = QAction(
            fonticon.icon(FA6S.paste),
            "&Paste",
            self,
            shortcut=QKeySequence.StandardKey.Paste,
            statusTip="Paste the clipboard's contents into the current selection",
            triggered=self._text_edit.paste,
        )

        self._about_act = QAction(
            "&About",
            self,
            statusTip="Show the application's About box",
            triggered=self.about,
        )

        self._cut_act.setEnabled(False)
        self._copy_act.setEnabled(False)
        self._text_edit.copyAvailable.connect(self._cut_act.setEnabled)
        self._text_edit.copyAvailable.connect(self._copy_act.setEnabled)

    def create_menus(self) -> None:
        self._file_menu = self.menuBar().addMenu("&File")
        self._file_menu.addAction(self._new_act)
        self._file_menu.addAction(self._open_act)
        self._file_menu.addAction(self._save_act)
        self._file_menu.addAction(self._save_as_act)
        self._file_menu.addSeparator()
        self._file_menu.addAction(self._exit_act)

        self._edit_menu = self.menuBar().addMenu("&Edit")
        self._edit_menu.addAction(self._cut_act)
        self._edit_menu.addAction(self._copy_act)
        self._edit_menu.addAction(self._paste_act)

        self.menuBar().addSeparator()

        self._help_menu = self.menuBar().addMenu("&Help")
        self._help_menu.addAction(self._about_act)

    def create_tool_bars(self) -> None:
        self._file_tool_bar = self.addToolBar("File")
        self._file_tool_bar.addAction(self._new_act)
        self._file_tool_bar.addAction(self._open_act)
        self._file_tool_bar.addAction(self._save_act)

        self._edit_tool_bar = self.addToolBar("Edit")
        self._edit_tool_bar.addAction(self._cut_act)
        self._edit_tool_bar.addAction(self._copy_act)
        self._edit_tool_bar.addAction(self._paste_act)

    def create_status_bar(self) -> None:
        self.statusBar().showMessage("Ready")

    def maybe_save(self) -> bool:
        if self._text_edit.document().isModified():
            ret = QMessageBox.warning(
                self,
                "Application",
                "The document has been modified.\nDo you want to save your changes?",
                QMessageBox.StandardButton.Save
                | QMessageBox.StandardButton.Discard
                | QMessageBox.StandardButton.Cancel,
            )
            if ret == QMessageBox.StandardButton.Save:
                return self.save()
            elif ret == QMessageBox.StandardButton.Cancel:
                return False
        return True

    def load_file(self, fileName: str) -> None:
        file = QFile(fileName)
        if not file.open(QFile.OpenModeFlag.ReadOnly | QFile.OpenModeFlag.Text):
            reason = file.errorString()
            QMessageBox.warning(
                self, "Application", f"Cannot read file {fileName}:\n{reason}."
            )
            return

        inf = QTextStream(file)
        QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
        self._text_edit.setPlainText(inf.readAll())
        QApplication.restoreOverrideCursor()

        self.set_current_file(fileName)
        self.statusBar().showMessage("File loaded", 2000)

    def save_file(self, fileName: str) -> bool:
        error = None
        QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
        file = QSaveFile(fileName)
        if file.open(QFile.OpenModeFlag.WriteOnly | QFile.OpenModeFlag.Text):
            outf = QTextStream(file)
            outf << self._text_edit.toPlainText()  # pyright: ignore
            if not file.commit():
                reason = file.errorString()
                error = f"Cannot write file {fileName}:\n{reason}."
        else:
            reason = file.errorString()
            error = f"Cannot open file {fileName}:\n{reason}."
        QApplication.restoreOverrideCursor()

        if error:
            QMessageBox.warning(self, "Application", error)
            return False

        self.set_current_file(fileName)
        self.statusBar().showMessage("File saved", 2000)
        return True

    def set_current_file(self, fileName: str) -> None:
        self._cur_file = fileName
        self._text_edit.document().setModified(False)
        self.setWindowModified(False)

        if self._cur_file:
            shown_name = self.stripped_name(self._cur_file)
        else:
            shown_name = "untitled.txt"

        self.setWindowTitle(f"{shown_name}[*] - Application")

    def stripped_name(self, fullFileName: str) -> str:
        return QFileInfo(fullFileName).fileName()


if __name__ == "__main__":
    qapp = QApplication.instance() or QApplication([])
    qapp.setAttribute(Qt.ApplicationAttribute.AA_DontShowIconsInMenus)
    main_win = MainWindow()
    main_win.show()
    qapp.exec()



import gc
from unittest.mock import Mock

import pytest

from app_model.expressions import Context, create_context, get_context
from app_model.expressions._context import _OBJ_TO_CONTEXT


def test_create_context() -> None:
    """You can create a context for any object"""

    class T: ...

    t = T()
    tid = id(t)
    ctx = create_context(t)
    assert get_context(t) == ctx
    assert hash(ctx)  # hashable
    assert tid in _OBJ_TO_CONTEXT
    _OBJ_TO_CONTEXT.pop(tid)

    del t
    gc.collect()
    assert tid not in _OBJ_TO_CONTEXT

    # you can provide your own root, but it must be a context
    create_context(T(), root=Context())
    with pytest.raises(AssertionError):
        create_context(T(), root={})  # type: ignore


def test_create_and_get_scoped_contexts() -> None:
    """Test that objects created in the stack of another contexted object.

    likely the most common way that this API will be used:
    """
    before = len(_OBJ_TO_CONTEXT)

    class A:
        def __init__(self) -> None:
            create_context(self)
            self.b = B()

    class B:
        def __init__(self) -> None:
            create_context(self)

    obja = A()
    assert len(_OBJ_TO_CONTEXT) == before + 2
    ctxa = get_context(obja)
    ctxb = get_context(obja.b)
    assert ctxa is not None
    assert ctxb is not None
    ctxa["hi"] = "hi"
    assert ctxb["hi"] == "hi"

    # keys get deleted on object deletion
    del obja
    gc.collect()
    assert len(_OBJ_TO_CONTEXT) == before


def test_context_events() -> None:
    """Changing context keys emits an event"""
    mock = Mock()
    root = Context()
    scoped = root.new_child()
    scoped.changed.connect(mock)  # connect the mock to the child

    root["a"] = 1
    # child re-emits parent events
    assert mock.call_args[0][0] == {"a"}

    mock.reset_mock()
    scoped["b"] = 1
    # also emits own events
    assert mock.call_args[0][0] == {"b"}

    mock.reset_mock()
    del scoped["b"]
    assert mock.call_args[0][0] == {"b"}

    # but parent does not emit child events
    mock.reset_mock()
    mock2 = Mock()
    root.changed.connect(mock2)
    scoped["c"] = "c"
    mock.assert_called_once()
    mock2.assert_not_called()

    mock.reset_mock()
    with scoped.buffered_changes():
        scoped["d"] = "d"
        scoped["e"] = "f"
        scoped["f"] = "f"
    mock.assert_called_once_with({"d", "e", "f"})

    # check events properly propagated when adding already existing context as child
    mock3 = Mock()
    root2a = Context()
    root2b = Context()
    scoped2 = root2a.new_child(root2b)
    scoped2.changed.connect(mock3)  # connect the mock to the child

    root2a["a"] = 1
    # child re-emits parent events
    assert mock3.call_args[0][0] == {"a"}

    mock3.reset_mock()
    root2b["b"] = 1
    # child re-emits added events
    assert mock3.call_args[0][0] == {"b"}

    mock3.reset_mock()
    scoped2["c"] = 1
    # also emits own events
    assert mock3.call_args[0][0] == {"c"}

    # check events properly propagated when making a context of contexts
    mock4 = Mock()
    root3a = Context()
    root3b = Context()
    root3c = Context()
    root3d = Context()
    root3e = Context()
    combined = Context(root3a, root3b, root3c, root3d, root3e)
    combined.changed.connect(mock4)

    root3a["a"] = 1
    assert mock4.call_args[0][0] == {"a"}

    mock4.reset_mock()
    root3b["b"] = 1
    assert mock4.call_args[0][0] == {"b"}

    mock4.reset_mock()
    root3c["c"] = 1
    assert mock4.call_args[0][0] == {"c"}

    mock4.reset_mock()
    root3d["d"] = 1
    assert mock4.call_args[0][0] == {"d"}

    mock4.reset_mock()
    root3e["e"] = 1
    assert mock4.call_args[0][0] == {"e"}



import pytest

from app_model.registries import CommandsRegistry, RegisteredCommand


def raise_exc() -> None:
    raise RuntimeError("boom")


def test_commands_registry() -> None:
    reg = CommandsRegistry()
    id1 = "my.id"
    reg.register_command(id1, lambda: 42, "My Title")
    assert "(1 commands)" in repr(reg)
    assert id1 in str(reg)
    assert id1 in reg
    with pytest.raises(KeyError, match=r"my\.id2"):
        reg["my.id2"]

    with pytest.raises(ValueError, match=r"Command 'my\.id' already registered"):
        reg.register_command(id1, lambda: 42, "My Title")

    assert reg.execute_command(id1, execute_asynchronously=True).result() == 42
    assert reg.execute_command(id1, execute_asynchronously=False).result() == 42

    reg.register_command("my.id2", raise_exc, "My Title 2")
    future_async = reg.execute_command("my.id2", execute_asynchronously=True)
    future_sync = reg.execute_command("my.id2", execute_asynchronously=False)

    with pytest.raises(RuntimeError, match="boom"):
        future_async.result()
    with pytest.raises(RuntimeError, match="boom"):
        future_sync.result()


def test_commands_raises() -> None:
    reg = CommandsRegistry(raise_synchronous_exceptions=True)

    id_ = "my.id"
    title = "My Title"
    reg.register_command(id_, raise_exc, title)

    with pytest.raises(RuntimeError, match="boom"):
        reg.execute_command(id_)

    cmd = reg[id_]
    assert isinstance(cmd, RegisteredCommand)
    assert cmd.title == title

    with pytest.raises(AttributeError, match="immutable"):
        cmd.title = "New Title"

    assert cmd.title == title



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from app_model import Action, Application
from app_model.types import KeyCode, KeyMod, SubmenuItem

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import NoReturn

try:
    from fonticon_fa6 import FA6S

    UNDO_ICON = FA6S.rotate_left
except ImportError:
    UNDO_ICON = "fa6-solid:undo"

FIXTURES = Path(__file__).parent / "fixtures"


class Menus:
    FILE = "file"
    EDIT = "edit"
    HELP = "help"
    FILE_OPEN_FROM = "file/open_from"


class Commands:
    TOP = "top"
    OPEN = "open"
    UNDO = "undo"
    REDO = "redo"
    COPY = "copy"
    PASTE = "paste"
    TOGGLE_THING = "toggle_thing"
    OPEN_FROM_A = "open.from_a"
    OPEN_FROM_B = "open.from_b"
    UNIMPORTABLE = "unimportable"
    NOT_CALLABLE = "not.callable"
    RAISES = "raises.error"


def _raise_an_error() -> NoReturn:
    raise ValueError("This is an error")


class Mocks:
    def __init__(self) -> None:
        self.open = Mock(name=Commands.OPEN)
        self.undo = Mock(name=Commands.UNDO)
        self.copy = Mock(name=Commands.COPY)
        self.paste = Mock(name=Commands.PASTE)
        self.open_from_a = Mock(name=Commands.OPEN_FROM_A)
        self.open_from_b = Mock(name=Commands.OPEN_FROM_B)

    @property
    def redo(self) -> Mock:
        """This tests that we can lazily import a callback.

        There is a function called `run_me` in fixtures/fake_module.py that calls the
        global mock in that module.  In the redo action below, we declare:
        `callback="fake_module:run_me"`

        So, whenever the redo action is triggered, it should import that module, and
        then call the mock.  We can also access it here at `mocks.redo`... but the
        fixtures directory must be added to sys path during the test (as we do below)
        """
        try:
            from fake_module import GLOBAL_MOCK

            return GLOBAL_MOCK
        except ImportError as e:
            raise ImportError(
                "This mock must be run with the fixutres directory added to sys.path."
            ) from e


class FullApp(Application):
    Menus = Menus
    Commands = Commands

    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.mocks = Mocks()


def build_app(name: str = "complete_test_app") -> FullApp:
    app = FullApp(name)
    app.menus.append_menu_items(
        [
            (
                Menus.FILE,
                SubmenuItem(
                    submenu=Menus.FILE_OPEN_FROM,
                    title="Open From...",
                    icon="fa6s.folder-open",
                    when="not something_open",
                    enablement="friday",
                ),
            )
        ]
    )

    actions: list[Action] = [
        Action(
            id=Commands.OPEN,
            title="Open...",
            callback=app.mocks.open,
            menus=[{"id": Menus.FILE}],
            keybindings=[{"primary": "Ctrl+O"}],
        ),
        # putting these above undo redo to make sure that group sorting works
        Action(
            id=Commands.COPY,
            title="Copy",
            icon="fa6-solid:copy",  # iconify font style works too
            callback=app.mocks.copy,
            menus=[{"id": Menus.EDIT, "group": "2_copy_paste"}],
            keybindings=[{"primary": KeyMod.CtrlCmd | KeyCode.KeyC}],
        ),
        Action(
            id=Commands.PASTE,
            title="Paste",
            icon="fa6s.paste",
            callback=app.mocks.paste,
            menus=[{"id": Menus.EDIT, "group": "2_copy_paste"}],
            keybindings=[{"primary": "Ctrl+V", "mac": "Cmd+V"}],
        ),
        # putting this above UNDO to make sure that order sorting works
        Action(
            id=Commands.REDO,
            title="Redo",
            tooltip="Redo it!",
            icon="fa6-solid:rotate-right",
            enablement="allow_undo_redo",
            callback="fake_module:run_me",  # this is a function in fixtures
            keybindings=[{"primary": "Ctrl+Shift+Z"}],
            menus=[
                {
                    "id": Menus.EDIT,
                    "group": "1_undo_redo",
                    "order": 1,
                    "when": "not something_to_undo",
                }
            ],
        ),
        Action(
            id=Commands.UNDO,
            tooltip="Undo it!",
            title="Undo",
            icon=UNDO_ICON,  # testing alternate way to specify icon
            enablement="allow_undo_redo",
            callback=app.mocks.undo,
            keybindings=[{"primary": "Ctrl+Z"}],
            menus=[
                {
                    "id": Menus.EDIT,
                    "group": "1_undo_redo",
                    "order": 0,
                    "when": "something_to_undo",
                }
            ],
        ),
        # test the navigation key
        Action(
            id=Commands.TOP,
            title="AtTop",
            callback=lambda: None,
            menus=[{"id": Menus.EDIT, "group": "navigation"}],
        ),
        # test submenus
        Action(
            id=Commands.OPEN_FROM_A,
            title="Open from A",
            callback=app.mocks.open_from_a,
            menus=[{"id": Menus.FILE_OPEN_FROM}],
        ),
        Action(
            id=Commands.OPEN_FROM_B,
            title="Open from B",
            callback=app.mocks.open_from_b,
            menus=[{"id": Menus.FILE_OPEN_FROM}],
            enablement="sat",
        ),
        Action(
            id=Commands.UNIMPORTABLE,
            title="Can't be found",
            callback="unresolvable:function",
        ),
        Action(
            id=Commands.NOT_CALLABLE,
            title="Will Never Work",
            callback="fake_module:attr",
        ),
        Action(
            id=Commands.RAISES,
            title="Will raise an error",
            callback=_raise_an_error,
        ),
        Action(
            id=Commands.TOGGLE_THING,
            title="Toggle Thing",
            callback=lambda: None,
            menus=[{"id": Menus.HELP}],
            toggled="thing_toggled",
        ),
    ]
    for action in actions:
        app.register_action(action)

    return app


@pytest.fixture
def full_app(monkeypatch: pytest.MonkeyPatch) -> Iterator[Application]:
    """Premade application."""
    try:
        app = build_app("complete_test_app")
        with monkeypatch.context() as m:
            # mock path to add fake_module
            m.setattr(sys, "path", [str(FIXTURES), *sys.path])
            # make sure it's not already in sys.modules
            sys.modules.pop("fake_module", None)
            yield app
            # clear the global mock if it's been called
            app.mocks.redo.reset_mock()
    finally:
        Application.destroy("complete_test_app")


@pytest.fixture
def simple_app() -> Iterator[Application]:
    app = Application("test")
    app.commands_changed = Mock()
    app.commands.registered.connect(app.commands_changed)
    app.keybindings_changed = Mock()
    app.keybindings.registered.connect(app.keybindings_changed)
    app.menus_changed = Mock()
    app.menus.menus_changed.connect(app.menus_changed)
    yield app
    Application.destroy("test")
    assert "test" not in Application._instances



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtCore import QUrl
from qtpy.QtGui import QIcon

if TYPE_CHECKING:
    from typing import Literal

    from app_model.types import Icon


def to_qicon(icon: Icon, theme: Literal["dark", "light"] = "dark") -> QIcon:
    """Create QIcon from Icon."""
    from superqt import QIconifyIcon, fonticon

    if icn := getattr(icon, theme, ""):
        if icn.startswith("file://"):
            return QIcon(QUrl(icn).toLocalFile())
        elif ":" in icn:
            return QIconifyIcon(icn)
        else:
            return fonticon.icon(icn)
    return QIcon()  # pragma: no cover



"""App-model types."""

from typing import TYPE_CHECKING

from ._action import Action
from ._command_rule import CommandRule, ToggleRule
from ._constants import KeyBindingSource, OperatingSystem
from ._icon import Icon
from ._keybinding_rule import KeyBindingRule
from ._keys import (
    KeyBinding,
    KeyChord,
    KeyCode,
    KeyCombo,
    KeyMod,
    ScanCode,
    SimpleKeyBinding,
    StandardKeyBinding,
)
from ._menu_rule import MenuItem, MenuItemBase, MenuRule, SubmenuItem

if TYPE_CHECKING:
    from typing import Callable

    from typing_extensions import TypeAlias

    from ._icon import IconOrDict as IconOrDict
    from ._keybinding_rule import KeyBindingRuleDict as KeyBindingRuleDict
    from ._keybinding_rule import KeyBindingRuleOrDict as KeyBindingRuleOrDict
    from ._menu_rule import MenuOrSubmenu as MenuOrSubmenu
    from ._menu_rule import MenuRuleDict as MenuRuleDict
    from ._menu_rule import MenuRuleOrDict as MenuRuleOrDict

    # function that can be called without arguments to dispose of a resource
    DisposeCallable: TypeAlias = Callable[[], None]


__all__ = [
    "Action",
    "CommandRule",
    "Icon",
    "KeyBinding",
    "KeyBindingRule",
    "KeyBindingSource",
    "KeyChord",
    "KeyCode",
    "KeyCombo",
    "KeyMod",
    "MenuItem",
    "MenuItemBase",
    "MenuRule",
    "OperatingSystem",
    "ScanCode",
    "SimpleKeyBinding",
    "StandardKeyBinding",
    "SubmenuItem",
    "ToggleRule",
]



from qtpy.QtGui import QKeySequence

from app_model.backends.qt import QModelKeyBindingEdit
from app_model.types import KeyBinding, KeyCode, KeyMod


def test_qkeysequenceedit(qtbot) -> None:
    edit = QModelKeyBindingEdit()
    qtbot.addWidget(edit)
    assert edit.keyBinding() is None
    edit.setKeySequence(QKeySequence("Shift+A"))
    assert edit.keyBinding() == KeyBinding(parts=[KeyMod.Shift | KeyCode.KeyA])



from typing import Callable

import pytest

from app_model.types._constants import OperatingSystem
from app_model.types._keys import KeyChord, KeyCode, KeyMod, ScanCode, SimpleKeyBinding
from app_model.types._keys._key_codes import keycode_to_os_name, keycode_to_os_symbol


def test_key_codes() -> None:
    for key in KeyCode:
        assert key == KeyCode.from_string(str(key))

    assert KeyCode.from_event_code(65) == KeyCode.KeyA

    assert KeyCode.validate(int(KeyCode.KeyA)) == KeyCode.KeyA
    assert KeyCode.validate(KeyCode.KeyA) == KeyCode.KeyA
    assert KeyCode.validate("A") == KeyCode.KeyA

    with pytest.raises(TypeError, match="cannot convert"):
        KeyCode.validate({"a"})


@pytest.mark.parametrize("symbol_or_name", ["symbol", "name"])
@pytest.mark.parametrize(
    ("os", "key_symbols_func", "key_names_func"),
    [
        (OperatingSystem.WINDOWS, keycode_to_os_symbol, keycode_to_os_name),
        (OperatingSystem.MACOS, keycode_to_os_symbol, keycode_to_os_name),
        (OperatingSystem.LINUX, keycode_to_os_symbol, keycode_to_os_name),
    ],
)
def test_key_codes_to_os(
    symbol_or_name: str,
    os: OperatingSystem,
    key_symbols_func: Callable[[KeyCode, OperatingSystem], str],
    key_names_func: Callable[[KeyCode, OperatingSystem], str],
) -> None:
    os_method = f"os_{symbol_or_name}"
    key_map_func = key_symbols_func if symbol_or_name == "symbol" else key_names_func
    for key in KeyCode:
        assert getattr(key, os_method)(os) == key_map_func(key, os)


def test_scan_codes() -> None:
    for scan in ScanCode:
        assert scan == ScanCode.from_string(str(scan)), scan


def test_key_combo() -> None:
    """KeyCombo is an integer combination of one or more KeyMod and KeyCode."""
    combo = KeyMod.Shift | KeyMod.Alt | KeyCode.KeyK
    assert repr(combo) == "<KeyCombo.Shift|Alt|KeyK: 1564>"
    kb = SimpleKeyBinding.from_int(combo)
    assert kb == SimpleKeyBinding(shift=True, alt=True, key=KeyCode.KeyK)


def test_key_chord() -> None:
    """KeyChord is an integer combination of two KeyCombos, KeyCodes, or integers."""
    chord = KeyChord(KeyMod.CtrlCmd | KeyCode.KeyK, KeyCode.KeyM)
    assert int(chord) == 1968156
    assert repr(chord) == "KeyChord(<KeyCombo.CtrlCmd|KeyK: 2076>, <KeyCode.KeyM: 30>)"



# mypy: disable-error-code="operator"
# pyright: reportOperatorIssue=false

from __future__ import annotations

import operator
from functools import reduce
from typing import TYPE_CHECKING

from qtpy.QtCore import QCoreApplication, Qt
from qtpy.QtGui import QKeySequence

from app_model.types import (
    KeyBinding,
    KeyCode,
    KeyCombo,
    KeyMod,
    SimpleKeyBinding,
)
from app_model.types._constants import OperatingSystem

if TYPE_CHECKING:
    from collections.abc import Mapping, MutableMapping

    from qtpy.QtCore import QKeyCombination

try:
    from qtpy import QT6
except ImportError:
    QT6 = False

QCTRL = Qt.KeyboardModifier.ControlModifier
QSHIFT = Qt.KeyboardModifier.ShiftModifier
QALT = Qt.KeyboardModifier.AltModifier
QMETA = Qt.KeyboardModifier.MetaModifier

MAC = OperatingSystem.current().is_mac

_QMOD_LOOKUP = {
    "ctrl": QCTRL,
    "shift": QSHIFT,
    "alt": QALT,
    "meta": QMETA,
}

_SWAPPED_QMOD_LOOKUP = {
    **_QMOD_LOOKUP,
    "ctrl": QMETA,
    "meta": QCTRL,
}


def _mac_ctrl_meta_swapped() -> bool:
    """Return True if Qt is swapping Ctrl and Meta for keyboard interactions."""
    return not QCoreApplication.testAttribute(
        Qt.ApplicationAttribute.AA_MacDontSwapCtrlAndMeta
    )


if QT6:
    from qtpy.QtCore import QKeyCombination

    def simple_keybinding_to_qint(skb: SimpleKeyBinding) -> int:
        """Create Qt Key integer from a SimpleKeyBinding."""
        lookup = (
            _SWAPPED_QMOD_LOOKUP if MAC and _mac_ctrl_meta_swapped() else _QMOD_LOOKUP
        )
        key = modelkey2qkey(skb.key) if skb.key else Qt.Key.Key_unknown
        mods = (v for k, v in lookup.items() if getattr(skb, k))
        combo = QKeyCombination(
            reduce(operator.or_, mods, Qt.KeyboardModifier.NoModifier), key
        )
        return int(combo.toCombined())

else:

    def simple_keybinding_to_qint(skb: SimpleKeyBinding) -> int:
        """Create Qt Key integer from a SimpleKeyBinding."""
        lookup = (
            _SWAPPED_QMOD_LOOKUP if MAC and _mac_ctrl_meta_swapped() else _QMOD_LOOKUP
        )
        out = modelkey2qkey(skb.key) if skb.key else 0
        mods = (v for k, v in lookup.items() if getattr(skb, k))
        out = reduce(operator.or_, mods, out)
        return int(out)


if QT6:
    # note: this doesn't work on pyside6 < 6.5 ...
    # but we don't support that anymore
    def _get_qmods(key: QKeyCombination) -> Qt.KeyboardModifier:
        return key.keyboardModifiers()

    def _get_qkey(key: QKeyCombination) -> Qt.Key:
        return key.key()

else:

    def _get_qmods(key: QKeyCombination) -> Qt.KeyboardModifier:
        return Qt.KeyboardModifier(key & Qt.KeyboardModifier.KeyboardModifierMask)

    def _get_qkey(key: QKeyCombination) -> Qt.Key:
        return Qt.Key(key & ~Qt.KeyboardModifier.KeyboardModifierMask)


# maybe ~ 1.5x faster than:
# QKeySequence.fromString(",".join(str(x) for x in kb.parts))
# but the string version might be more reliable?
class QKeyBindingSequence(QKeySequence):
    """A QKeySequence based on a KeyBinding instance."""

    def __init__(self, kb: KeyBinding) -> None:
        ints = [simple_keybinding_to_qint(skb) for skb in kb.parts]
        super().__init__(*ints)


KEY_TO_QT: dict[KeyCode | None, Qt.Key] = {
    None: Qt.Key.Key_unknown,
    KeyCode.UNKNOWN: Qt.Key.Key_unknown,
    KeyCode.Backquote: Qt.Key.Key_QuoteLeft,
    KeyCode.Backslash: Qt.Key.Key_Backslash,
    KeyCode.IntlBackslash: Qt.Key.Key_Backslash,
    KeyCode.BracketLeft: Qt.Key.Key_BracketLeft,
    KeyCode.BracketRight: Qt.Key.Key_BracketRight,
    KeyCode.Comma: Qt.Key.Key_Comma,
    KeyCode.Digit0: Qt.Key.Key_0,
    KeyCode.Digit1: Qt.Key.Key_1,
    KeyCode.Digit2: Qt.Key.Key_2,
    KeyCode.Digit3: Qt.Key.Key_3,
    KeyCode.Digit4: Qt.Key.Key_4,
    KeyCode.Digit5: Qt.Key.Key_5,
    KeyCode.Digit6: Qt.Key.Key_6,
    KeyCode.Digit7: Qt.Key.Key_7,
    KeyCode.Digit8: Qt.Key.Key_8,
    KeyCode.Digit9: Qt.Key.Key_9,
    KeyCode.Equal: Qt.Key.Key_Equal,
    KeyCode.KeyA: Qt.Key.Key_A,
    KeyCode.KeyB: Qt.Key.Key_B,
    KeyCode.KeyC: Qt.Key.Key_C,
    KeyCode.KeyD: Qt.Key.Key_D,
    KeyCode.KeyE: Qt.Key.Key_E,
    KeyCode.KeyF: Qt.Key.Key_F,
    KeyCode.KeyG: Qt.Key.Key_G,
    KeyCode.KeyH: Qt.Key.Key_H,
    KeyCode.KeyI: Qt.Key.Key_I,
    KeyCode.KeyJ: Qt.Key.Key_J,
    KeyCode.KeyK: Qt.Key.Key_K,
    KeyCode.KeyL: Qt.Key.Key_L,
    KeyCode.KeyM: Qt.Key.Key_M,
    KeyCode.KeyN: Qt.Key.Key_N,
    KeyCode.KeyO: Qt.Key.Key_O,
    KeyCode.KeyP: Qt.Key.Key_P,
    KeyCode.KeyQ: Qt.Key.Key_Q,
    KeyCode.KeyR: Qt.Key.Key_R,
    KeyCode.KeyS: Qt.Key.Key_S,
    KeyCode.KeyT: Qt.Key.Key_T,
    KeyCode.KeyU: Qt.Key.Key_U,
    KeyCode.KeyV: Qt.Key.Key_V,
    KeyCode.KeyW: Qt.Key.Key_W,
    KeyCode.KeyX: Qt.Key.Key_X,
    KeyCode.KeyY: Qt.Key.Key_Y,
    KeyCode.KeyZ: Qt.Key.Key_Z,
    KeyCode.Minus: Qt.Key.Key_Minus,
    KeyCode.Period: Qt.Key.Key_Period,
    KeyCode.Quote: Qt.Key.Key_Apostrophe,
    KeyCode.Semicolon: Qt.Key.Key_Semicolon,
    KeyCode.Slash: Qt.Key.Key_Slash,
    KeyCode.Alt: Qt.Key.Key_Alt,
    KeyCode.Backspace: Qt.Key.Key_Backspace,
    KeyCode.CapsLock: Qt.Key.Key_CapsLock,
    KeyCode.ContextMenu: Qt.Key.Key_Context1,
    KeyCode.Ctrl: Qt.Key.Key_Control,
    KeyCode.Enter: Qt.Key.Key_Enter,
    KeyCode.Meta: Qt.Key.Key_Meta,
    KeyCode.Shift: Qt.Key.Key_Shift,
    KeyCode.Space: Qt.Key.Key_Space,
    KeyCode.Tab: Qt.Key.Key_Tab,
    KeyCode.Delete: Qt.Key.Key_Delete,
    KeyCode.End: Qt.Key.Key_End,
    KeyCode.Home: Qt.Key.Key_Home,
    KeyCode.Insert: Qt.Key.Key_Insert,
    KeyCode.PageDown: Qt.Key.Key_PageDown,
    KeyCode.PageUp: Qt.Key.Key_PageUp,
    KeyCode.DownArrow: Qt.Key.Key_Down,
    KeyCode.LeftArrow: Qt.Key.Key_Left,
    KeyCode.RightArrow: Qt.Key.Key_Right,
    KeyCode.UpArrow: Qt.Key.Key_Up,
    KeyCode.NumLock: Qt.Key.Key_NumLock,
    KeyCode.Numpad0: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_0,
    KeyCode.Numpad1: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_1,
    KeyCode.Numpad2: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_2,
    KeyCode.Numpad3: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_3,
    KeyCode.Numpad4: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_4,
    KeyCode.Numpad5: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_5,
    KeyCode.Numpad6: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_6,
    KeyCode.Numpad7: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_7,
    KeyCode.Numpad8: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_8,
    KeyCode.Numpad9: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_9,
    KeyCode.NumpadAdd: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_Plus,
    KeyCode.NumpadDecimal: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_Period,
    KeyCode.NumpadDivide: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_Slash,
    KeyCode.NumpadMultiply: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_Asterisk,
    KeyCode.NumpadSubtract: Qt.KeyboardModifier.KeypadModifier | Qt.Key.Key_Minus,
    KeyCode.Escape: Qt.Key.Key_Escape,
    KeyCode.F1: Qt.Key.Key_F1,
    KeyCode.F2: Qt.Key.Key_F2,
    KeyCode.F3: Qt.Key.Key_F3,
    KeyCode.F4: Qt.Key.Key_F4,
    KeyCode.F5: Qt.Key.Key_F5,
    KeyCode.F6: Qt.Key.Key_F6,
    KeyCode.F7: Qt.Key.Key_F7,
    KeyCode.F8: Qt.Key.Key_F8,
    KeyCode.F9: Qt.Key.Key_F9,
    KeyCode.F10: Qt.Key.Key_F10,
    KeyCode.F11: Qt.Key.Key_F11,
    KeyCode.F12: Qt.Key.Key_F12,
    KeyCode.PrintScreen: Qt.Key.Key_Print,
    KeyCode.ScrollLock: Qt.Key.Key_ScrollLock,
    KeyCode.PauseBreak: Qt.Key.Key_Pause,
}

KEYMOD_FROM_QT = {
    Qt.KeyboardModifier.NoModifier: KeyMod.NONE,
    QALT: KeyMod.Alt,
    QCTRL: KeyMod.CtrlCmd,
    QSHIFT: KeyMod.Shift,
    QMETA: KeyMod.WinCtrl,
}

MAC_KEYMOD_FROM_QT = {**KEYMOD_FROM_QT, QCTRL: KeyMod.WinCtrl, QMETA: KeyMod.CtrlCmd}

KEYMOD_TO_QT = {
    KeyMod.NONE: Qt.KeyboardModifier.NoModifier,
    KeyMod.CtrlCmd: QCTRL,
    KeyMod.Alt: QALT,
    KeyMod.Shift: QSHIFT,
    KeyMod.WinCtrl: QMETA,
}

MAC_KEYMOD_TO_QT = {**KEYMOD_TO_QT, KeyMod.WinCtrl: QCTRL, KeyMod.CtrlCmd: QMETA}


KEY_FROM_QT: MutableMapping[Qt.Key, KeyCode | KeyCombo] = {
    v.toCombined() if hasattr(v, "toCombined") else int(v): k  # pyright: ignore
    for k, v in KEY_TO_QT.items()
    if k
}

# Qt Keys which have no representation in the W3C spec
_QTONLY_KEYS: Mapping[Qt.Key, KeyCode | KeyCombo] = {
    Qt.Key.Key_Exclam: KeyMod.Shift | KeyCode.Digit1,
    Qt.Key.Key_At: KeyMod.Shift | KeyCode.Digit2,
    Qt.Key.Key_NumberSign: KeyMod.Shift | KeyCode.Digit3,
    Qt.Key.Key_Dollar: KeyMod.Shift | KeyCode.Digit4,
    Qt.Key.Key_Percent: KeyMod.Shift | KeyCode.Digit5,
    Qt.Key.Key_AsciiCircum: KeyMod.Shift | KeyCode.Digit6,
    Qt.Key.Key_Ampersand: KeyMod.Shift | KeyCode.Digit7,
    Qt.Key.Key_Asterisk: KeyMod.Shift | KeyCode.Digit8,
    Qt.Key.Key_ParenLeft: KeyMod.Shift | KeyCode.Digit9,
    Qt.Key.Key_ParenRight: KeyMod.Shift | KeyCode.Digit0,
    Qt.Key.Key_Underscore: KeyMod.Shift | KeyCode.Minus,
    Qt.Key.Key_Plus: KeyMod.Shift | KeyCode.Equal,
    Qt.Key.Key_BraceLeft: KeyMod.Shift | KeyCode.BracketLeft,
    Qt.Key.Key_BraceRight: KeyMod.Shift | KeyCode.BracketRight,
    Qt.Key.Key_Bar: KeyMod.Shift | KeyCode.Backslash,
    Qt.Key.Key_Colon: KeyMod.Shift | KeyCode.Semicolon,
    Qt.Key.Key_QuoteDbl: KeyMod.Shift | KeyCode.Quote,
    Qt.Key.Key_Less: KeyMod.Shift | KeyCode.Comma,
    Qt.Key.Key_Greater: KeyMod.Shift | KeyCode.Period,
    Qt.Key.Key_Question: KeyMod.Shift | KeyCode.Slash,
    Qt.Key.Key_AsciiTilde: KeyMod.Shift | KeyCode.Backquote,
    Qt.Key.Key_Return: KeyCode.Enter,
    Qt.Key.Key_Backtab: KeyMod.Shift | KeyCode.Tab,
}
KEY_FROM_QT.update(_QTONLY_KEYS)


def qmods2modelmods(modifiers: Qt.KeyboardModifier) -> KeyMod:
    """Return KeyMod from Qt.KeyboardModifier."""
    mod = KeyMod.NONE
    lookup = (
        MAC_KEYMOD_FROM_QT if MAC and not _mac_ctrl_meta_swapped() else KEYMOD_FROM_QT
    )
    for modifier in lookup:
        if modifiers & modifier:
            mod |= lookup[modifier]
    return mod


def modelkey2qkey(key: KeyCode) -> Qt.Key:
    """Return Qt.Key from KeyCode."""
    if MAC and _mac_ctrl_meta_swapped():
        if key == KeyCode.Meta:
            return Qt.Key.Key_Control
        if key == KeyCode.Ctrl:
            return Qt.Key.Key_Meta
    return KEY_TO_QT.get(key, Qt.Key.Key_unknown)


def qkey2modelkey(key: Qt.Key) -> KeyCode | KeyCombo:
    """Return KeyCode from Qt.Key."""
    if MAC and _mac_ctrl_meta_swapped():
        if key == Qt.Key.Key_Control:
            return KeyCode.Meta
        if key == Qt.Key.Key_Meta:
            return KeyCode.Ctrl
    return KEY_FROM_QT.get(key, KeyCode.UNKNOWN)


def qkeycombo2modelkey(key: QKeyCombination) -> KeyCode | KeyCombo:
    """Return KeyCode or KeyCombo from QKeyCombination."""
    if key in KEY_FROM_QT:
        # type ignore because in qt5, key may actually just be int ... but it's fine.
        return KEY_FROM_QT[key]
    qmods = _get_qmods(key)
    qkey = _get_qkey(key)
    return qmods2modelmods(qmods) | qkey2modelkey(qkey)  # type: ignore [return-value]


def qkeysequence2modelkeybinding(key: QKeySequence) -> KeyBinding:
    """Return KeyBinding from QKeySequence."""
    # FIXME: this should return KeyChord instead of KeyBinding... but that only takes 2
    parts = [SimpleKeyBinding.from_int(qkeycombo2modelkey(x)) for x in iter(key)]
    return KeyBinding(parts=parts)


# ################# These are the Qkeys we currently aren't mapping ################ #
# Key_F14
# Key_F15
# Key_F16
# Key_F17
# Key_F18
# Key_F19
# Key_F20
# Key_F21
# Key_F22
# Key_F23
# Key_F24
# Key_F25
# Key_F26
# Key_F27
# Key_F28
# Key_F29
# Key_F30
# Key_F31
# Key_F32
# Key_F33
# Key_F34
# Key_F35
# Key_Super_L
# Key_Super_R
# Key_Menu
# Key_Hyper_L
# Key_Hyper_R
# Key_Help
# Key_Direction_L
# Key_Direction_R
# Key_nobreakspace
# Key_exclamdown
# Key_cent
# Key_sterling
# Key_currency
# Key_yen
# Key_brokenbar
# Key_section
# Key_diaeresis
# Key_copyright
# Key_ordfeminine
# Key_guillemotleft
# Key_notsign
# Key_hyphen
# Key_registered
# Key_macron
# Key_degree
# Key_plusminus
# Key_twosuperior
# Key_threesuperior
# Key_acute
# Key_mu
# Key_paragraph
# Key_periodcentered
# Key_cedilla
# Key_onesuperior
# Key_masculine
# Key_guillemotright
# Key_onequarter
# Key_onehalf
# Key_threequarters
# Key_questiondown
# Key_Agrave
# Key_Aacute
# Key_Acircumflex
# Key_Atilde
# Key_Adiaeresis
# Key_Aring
# Key_AE
# Key_Ccedilla
# Key_Egrave
# Key_Eacute
# Key_Ecircumflex
# Key_Ediaeresis
# Key_Igrave
# Key_Iacute
# Key_Icircumflex
# Key_Idiaeresis
# Key_ETH
# Key_Ntilde
# Key_Ograve
# Key_Oacute
# Key_Ocircumflex
# Key_Otilde
# Key_Odiaeresis
# Key_multiply
# Key_Ooblique
# Key_Ugrave
# Key_Uacute
# Key_Ucircumflex
# Key_Udiaeresis
# Key_Yacute
# Key_THORN
# Key_ssharp
# Key_division
# Key_ydiaeresis
# Key_AltGr
# Key_Multi_key
# Key_Codeinput
# Key_SingleCandidate
# Key_MultipleCandidate
# Key_PreviousCandidate
# Key_Mode_switch
# Key_Kanji
# Key_Muhenkan
# Key_Henkan
# Key_Romaji
# Key_Hiragana
# Key_Katakana
# Key_Hiragana_Katakana
# Key_Zenkaku
# Key_Hankaku
# Key_Zenkaku_Hankaku
# Key_Touroku
# Key_Massyo
# Key_Kana_Lock
# Key_Kana_Shift
# Key_Eisu_Shift
# Key_Eisu_toggle
# Key_Hangul
# Key_Hangul_Start
# Key_Hangul_End
# Key_Hangul_Hanja
# Key_Hangul_Jamo
# Key_Hangul_Romaja
# Key_Hangul_Jeonja
# Key_Hangul_Banja
# Key_Hangul_PreHanja
# Key_Hangul_PostHanja
# Key_Hangul_Special
# Key_Dead_Grave
# Key_Dead_Acute
# Key_Dead_Circumflex
# Key_Dead_Tilde
# Key_Dead_Macron
# Key_Dead_Breve
# Key_Dead_Abovedot
# Key_Dead_Diaeresis
# Key_Dead_Abovering
# Key_Dead_Doubleacute
# Key_Dead_Caron
# Key_Dead_Cedilla
# Key_Dead_Ogonek
# Key_Dead_Iota
# Key_Dead_Voiced_Sound
# Key_Dead_Semivoiced_Sound
# Key_Dead_Belowdot
# Key_Dead_Hook
# Key_Dead_Horn
# Key_Dead_Stroke
# Key_Dead_Abovecomma
# Key_Dead_Abovereversedcomma
# Key_Dead_Doublegrave
# Key_Dead_Belowring
# Key_Dead_Belowmacron
# Key_Dead_Belowcircumflex
# Key_Dead_Belowtilde
# Key_Dead_Belowbreve
# Key_Dead_Belowdiaeresis
# Key_Dead_Invertedbreve
# Key_Dead_Belowcomma
# Key_Dead_Currency
# Key_Dead_a
# Key_Dead_A
# Key_Dead_e
# Key_Dead_E
# Key_Dead_i
# Key_Dead_I
# Key_Dead_o
# Key_Dead_O
# Key_Dead_u
# Key_Dead_U
# Key_Dead_Small_Schwa
# Key_Dead_Capital_Schwa
# Key_Dead_Greek
# Key_Dead_Lowline
# Key_Dead_Aboveverticalline
# Key_Dead_Belowverticalline
# Key_Dead_Longsolidusoverlay
# Key_Back
# Key_Forward
# Key_Stop
# Key_Refresh
# Key_VolumeDown
# Key_VolumeMute
# Key_VolumeUp
# Key_BassBoost
# Key_BassUp
# Key_BassDown
# Key_TrebleUp
# Key_TrebleDown
# Key_MediaPlay
# Key_MediaStop
# Key_MediaPrevious
# Key_MediaNext
# Key_MediaRecord
# Key_MediaPause
# Key_MediaTogglePlayPause
# Key_HomePage
# Key_Favorites
# Key_Search
# Key_Standby
# Key_OpenUrl
# Key_LaunchMail
# Key_LaunchMedia
# Key_Launch0
# Key_Launch1
# Key_Launch2
# Key_Launch3
# Key_Launch4
# Key_Launch5
# Key_Launch6
# Key_Launch7
# Key_Launch8
# Key_Launch9
# Key_LaunchA
# Key_LaunchB
# Key_LaunchC
# Key_LaunchD
# Key_LaunchE
# Key_LaunchF
# Key_MonBrightnessUp
# Key_MonBrightnessDown
# Key_KeyboardLightOnOff
# Key_KeyboardBrightnessUp
# Key_KeyboardBrightnessDown
# Key_PowerOff
# Key_WakeUp
# Key_Eject
# Key_ScreenSaver
# Key_WWW
# Key_Memo
# Key_LightBulb
# Key_Shop
# Key_History
# Key_AddFavorite
# Key_HotLinks
# Key_BrightnessAdjust
# Key_Finance
# Key_Community
# Key_AudioRewind
# Key_BackForward
# Key_ApplicationLeft
# Key_ApplicationRight
# Key_Book
# Key_CD
# Key_Calculator
# Key_ToDoList
# Key_ClearGrab
# Key_Close
# Key_Copy
# Key_Cut
# Key_Display
# Key_DOS
# Key_Documents
# Key_Excel
# Key_Explorer
# Key_Game
# Key_Go
# Key_iTouch
# Key_LogOff
# Key_Market
# Key_Meeting
# Key_MenuKB
# Key_MenuPB
# Key_MySites
# Key_News
# Key_OfficeHome
# Key_Option
# Key_Paste
# Key_Phone
# Key_Calendar
# Key_Reply
# Key_Reload
# Key_RotateWindows
# Key_RotationPB
# Key_RotationKB
# Key_Save
# Key_Send
# Key_Spell
# Key_SplitScreen
# Key_Support
# Key_TaskPane
# Key_Terminal
# Key_Tools
# Key_Travel
# Key_Video
# Key_Word
# Key_Xfer
# Key_ZoomIn
# Key_ZoomOut
# Key_Away
# Key_Messenger
# Key_WebCam
# Key_MailForward
# Key_Pictures
# Key_Music
# Key_Battery
# Key_Bluetooth
# Key_WLAN
# Key_UWB
# Key_AudioForward
# Key_AudioRepeat
# Key_AudioRandomPlay
# Key_Subtitle
# Key_AudioCycleTrack
# Key_Time
# Key_Hibernate
# Key_View
# Key_TopMenu
# Key_PowerDown
# Key_Suspend
# Key_ContrastAdjust
# Key_LaunchG
# Key_LaunchH
# Key_TouchpadToggle
# Key_TouchpadOn
# Key_TouchpadOff
# Key_MicMute
# Key_Red
# Key_Green
# Key_Yellow
# Key_Blue
# Key_ChannelUp
# Key_ChannelDown
# Key_Guide
# Key_Info
# Key_Settings
# Key_MicVolumeUp
# Key_MicVolumeDown
# Key_New
# Key_Open
# Key_Find
# Key_Undo
# Key_Redo
# Key_MediaLast
# Key_Select
# Key_Yes
# Key_No
# Key_Cancel
# Key_Printer
# Key_Execute
# Key_Sleep
# Key_Play
# Key_Zoom
# Key_Exit
# Key_Context2
# Key_Context3
# Key_Context4
# Key_Call
# Key_Hangup
# Key_Flip
# Key_ToggleCallHangup
# Key_VoiceDial
# Key_LastNumberRedial
# Key_Camera
# Key_CameraFocus



from typing import TYPE_CHECKING, ClassVar

from pydantic import BaseModel

if TYPE_CHECKING:
    from pydantic import ConfigDict


class _BaseModel(BaseModel):
    """Base model for all types."""

    # don't switch to exclude ... it makes it hard to add fields to the
    # schema without breaking backwards compatibility
    model_config: ClassVar["ConfigDict"] = {"frozen": True, "extra": "ignore"}



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QAction, QApplication

from app_model.backends.qt import QModelMainWindow, QModelToolBar
from app_model.types._action import Action

if TYPE_CHECKING:
    from pytestqt.qtbot import QtBot

    from ..conftest import FullApp  # noqa: TID252


def test_qmodel_main_window(
    qtbot: QtBot, qapp: QApplication, full_app: FullApp
) -> None:
    win = QModelMainWindow(full_app)
    qtbot.addWidget(win)

    win.setModelMenuBar(
        {
            full_app.Menus.FILE: "File",
            full_app.Menus.EDIT: "Edit",
            full_app.Menus.HELP: "Help",
        }
    )
    assert [a.text() for a in win.menuBar().actions()] == ["File", "Edit", "Help"]

    tb = win.addModelToolBar(
        full_app.Menus.FILE,
        toolbutton_style=Qt.ToolButtonStyle.ToolButtonTextBesideIcon,
    )
    assert isinstance(tb, QModelToolBar)
    win.addModelToolBar(full_app.Menus.EDIT, area=Qt.ToolBarArea.RightToolBarArea)

    full_app.register_action(
        Action(
            id="late-action",
            title="Late Action",
            keybindings=[{"primary": "Shift+L"}],
            menus=[{"id": full_app.Menus.FILE}],
            callback=lambda: None,
        )
    )

    action = qapp.findChild(QAction, "late-action")
    assert action.shortcut().toString() == "Shift+L"



from typing import Callable, Optional, Union

from pydantic import Field

from app_model import expressions

from ._base import _BaseModel
from ._icon import Icon


class ToggleRule(_BaseModel):
    """More detailed description of a toggle rule."""

    condition: Optional[expressions.Expr] = Field(
        default=None,
        description="(Optional) Condition under which the command should appear "
        "checked/toggled in any GUI representation (like a menu or button).",
    )
    get_current: Optional[Callable[[], bool]] = Field(
        default=None,
        description="Function that returns the current state of the toggle.",
    )


class CommandRule(_BaseModel):
    """Data representing a command and its presentation.

    Presentation of contributed commands depends on the containing menu. The Command
    Palette, for instance, prefixes commands with their category, allowing for easy
    grouping. However, the Command Palette doesn't show icons nor disabled commands.
    Menus, on the other hand, shows disabled items as grayed out, but don't show the
    category label.
    """

    id: str = Field(..., description="A global identifier for the command.")
    title: str = Field(
        ...,
        description="Title by which the command is represented in the UI.",
    )
    category: Optional[str] = Field(
        default=None,
        description="(Optional) Category string by which the command may be grouped "
        "in the UI",
    )
    tooltip: Optional[str] = Field(
        default=None, description="(Optional) Tooltip to show when hovered."
    )
    status_tip: Optional[str] = Field(
        default=None,
        description="(Optional) Help message to show in the status bar when a "
        "button representing this command is hovered (for backends that support it).",
    )
    icon: Optional[Icon] = Field(
        default=None,
        description="(Optional) Icon used to represent this command, e.g. on buttons "
        "or in menus. These may be [iconify keys](https://icon-sets.iconify.design), "
        "such as `fa6-solid:arrow-down`, or "
        "[superqt.fonticon](https://pyapp-kit.github.io/superqt/utilities/fonticon/)"
        " keys, such as `fa6s.arrow_down`, or a path to a local `.svg` file using the "
        "[file URI scheme](https://en.wikipedia.org/wiki/File_URI_scheme). "
        "Note that on Windows the file URI scheme should always start with "
        "`file:///` (three slashes)",
    )
    icon_visible_in_menu: bool = Field(
        default=True,
        description="Whether to show the icon in menus (for backends that support it). "
        "If `False`, only the title will be shown. By default, `True`.",
    )
    enablement: Optional[expressions.Expr] = Field(
        default=None,
        description="(Optional) Condition which must be true to enable the command in "
        "the UI (menu and keybindings). Does not prevent executing the command by "
        "other means, like the `execute_command` API.",
    )
    short_title: Optional[str] = Field(
        default=None,
        description="(Optional) Short title by which the command is represented in "
        "the UI. Menus pick either `title` or `short_title` depending on the context "
        "in which they show commands.",
    )
    toggled: Union[ToggleRule, expressions.Expr, None] = Field(
        default=None,
        description="(Optional) Condition under which the command should appear "
        "checked/toggled in any GUI representation (like a menu or button).",
    )

    def _as_command_rule(self) -> "CommandRule":
        """Simplify (subclasses) to a plain CommandRule."""
        return CommandRule(**{f: getattr(self, f) for f in CommandRule.__annotations__})



from pathlib import Path
from typing import TYPE_CHECKING, cast

from qtpy.QtCore import QFile, QFileInfo, QSaveFile, Qt, QTextStream
from qtpy.QtWidgets import QApplication, QFileDialog, QMessageBox, QTextEdit

from app_model import Application, types
from app_model.backends.qt import QModelMainWindow
from app_model.expressions import create_context

if TYPE_CHECKING:
    from app_model.backends.qt._qmenu import QModelMenuBar


class MainWindow(QModelMainWindow):
    def __init__(self, app: Application) -> None:
        super().__init__(app)

        self._cur_file: str = ""
        self._text_edit = QTextEdit()
        self._text_edit.copyAvailable.connect(self._update_context)

        self.setCentralWidget(self._text_edit)
        self.setModelMenuBar([MenuId.FILE, MenuId.EDIT, MenuId.HELP])
        self.addModelToolBar(MenuId.FILE, exclude={CommandId.SAVE_AS, CommandId.EXIT})
        self.addModelToolBar(MenuId.EDIT)
        self.addModelToolBar(MenuId.HELP)
        if sb := self.statusBar():
            sb.showMessage("Ready")

        self.set_current_file("")

        self._ctx = create_context(self)
        self._ctx.changed.connect(self._on_context_changed)
        self._ctx["copyAvailable"] = False

    def _update_context(self, available: bool) -> None:
        self._ctx["copyAvailable"] = available

    def _on_context_changed(self) -> None:
        mb = cast("QModelMenuBar", self.menuBar())
        mb.update_from_context(self._ctx)

    def set_current_file(self, fileName: str) -> None:
        self._cur_file = fileName
        if doc := self._text_edit.document():
            doc.setModified(False)
        self.setWindowModified(False)

        if self._cur_file:
            shown_name = QFileInfo(self._cur_file).fileName()
        else:
            shown_name = "untitled.txt"

        self.setWindowTitle(f"{shown_name}[*] - Application")

    def save(self) -> bool:
        return self.save_file(self._cur_file) if self._cur_file else self.save_as()

    def save_as(self) -> bool:
        fileName, _ = QFileDialog.getSaveFileName(self)
        if fileName:
            return self.save_file(fileName)
        return False

    def save_file(self, fileName: str) -> bool:
        error = None
        QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
        file = QSaveFile(fileName)
        if file.open(QFile.OpenModeFlag.WriteOnly | QFile.OpenModeFlag.Text):  # type: ignore
            outf = QTextStream(file)
            outf << self._text_edit.toPlainText()  # pyright: ignore
            if not file.commit():
                reason = file.errorString()
                error = f"Cannot write file {fileName}:\n{reason}."
        else:
            reason = file.errorString()
            error = f"Cannot open file {fileName}:\n{reason}."
        QApplication.restoreOverrideCursor()

        if error:
            QMessageBox.warning(self, "Application", error)
            return False
        return True

    def maybe_save(self) -> bool:
        if (doc := self._text_edit.document()) and doc.isModified():
            ret = QMessageBox.warning(
                self,
                "Application",
                "The document has been modified.\nDo you want to save your changes?",
                QMessageBox.StandardButton.Save  # type: ignore
                | QMessageBox.StandardButton.Discard
                | QMessageBox.StandardButton.Cancel,
            )
            if ret == QMessageBox.StandardButton.Save:
                return self.save()
            elif ret == QMessageBox.StandardButton.Cancel:
                return False
        return True

    def new_file(self) -> None:
        if self.maybe_save():
            self._text_edit.clear()
            self.set_current_file("")

    def open_file(self) -> None:
        if self.maybe_save():
            fileName, _ = QFileDialog.getOpenFileName(self)
            if fileName:
                self.load_file(fileName)

    def load_file(self, fileName: str) -> None:
        file = QFile(fileName)
        if not file.open(QFile.OpenModeFlag.ReadOnly | QFile.OpenModeFlag.Text):  # type: ignore
            reason = file.errorString()
            QMessageBox.warning(
                self, "Application", f"Cannot read file {fileName}:\n{reason}."
            )
            return

        inf = QTextStream(file)
        QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
        self._text_edit.setPlainText(inf.readAll())
        QApplication.restoreOverrideCursor()

        self.set_current_file(fileName)
        if sb := self.statusBar():
            sb.showMessage("File loaded", 2000)

    def about(self) -> None:
        QMessageBox.about(
            self,
            "About Application",
            "The <b>Application</b> example demonstrates how to write "
            "modern GUI applications using Qt, with a menu bar, "
            "toolbars, and a status bar.",
        )

    def cut(self) -> None:
        self._text_edit.cut()

    def copy(self) -> None:
        self._text_edit.copy()

    def paste(self) -> None:
        self._text_edit.paste()

    def close(self) -> bool:
        return super().close()


# Actions defined declaratively outside of QMainWindow class ...
# menus and toolbars will be made and added automatically


class MenuId:
    FILE = "file"
    EDIT = "edit"
    HELP = "help"


class CommandId:
    SAVE_AS = "save_file_as"
    EXIT = "exit"


ABOUT_ICON_PATH = Path(__file__).parent / "images" / "about.svg"

ACTIONS: list[types.Action] = [
    types.Action(
        id="new_file",
        icon="fa6-solid:file-circle-plus",
        title="New",
        keybindings=[types.StandardKeyBinding.New],
        status_tip="Create a new file",
        menus=[{"id": MenuId.FILE, "group": "1_loadsave"}],
        callback=MainWindow.new_file,
    ),
    types.Action(
        id="open_file",
        icon="fa6-solid:folder-open",
        title="Open...",
        keybindings=[types.StandardKeyBinding.Open],
        status_tip="Open an existing file",
        menus=[{"id": MenuId.FILE, "group": "1_loadsave"}],
        callback=MainWindow.open_file,
    ),
    types.Action(
        id="save_file",
        icon="fa6-solid:floppy-disk",
        title="Save",
        keybindings=[types.StandardKeyBinding.Save],
        status_tip="Save the document to disk",
        menus=[{"id": MenuId.FILE, "group": "1_loadsave"}],
        callback=MainWindow.save,
    ),
    types.Action(
        id=CommandId.SAVE_AS,
        title="Save As...",
        keybindings=[types.StandardKeyBinding.SaveAs],
        status_tip="Save the document under a new name",
        menus=[{"id": MenuId.FILE, "group": "1_loadsave"}],
        callback=MainWindow.save_as,
    ),
    types.Action(
        id=CommandId.EXIT,
        title="Exit",
        keybindings=[types.StandardKeyBinding.Quit],
        status_tip="Exit the application",
        menus=[{"id": MenuId.FILE, "group": "3_launchexit"}],
        callback=MainWindow.close,
    ),
    types.Action(
        id="cut",
        icon="fa6-solid:scissors",
        title="Cut",
        keybindings=[types.StandardKeyBinding.Cut],
        enablement="copyAvailable",
        status_tip="Cut the current selection's contents to the clipboard",
        menus=[{"id": MenuId.EDIT}],
        callback=MainWindow.cut,
    ),
    types.Action(
        id="copy",
        icon="fa6-solid:copy",
        title="Copy",
        keybindings=[types.StandardKeyBinding.Copy],
        enablement="copyAvailable",
        status_tip="Copy the current selection's contents to the clipboard",
        menus=[{"id": MenuId.EDIT}],
        callback=MainWindow.copy,
    ),
    types.Action(
        id="paste",
        icon="fa6-solid:paste",
        title="Paste",
        keybindings=[types.StandardKeyBinding.Paste],
        status_tip="Paste the clipboard's contents into the current selection",
        menus=[{"id": MenuId.EDIT}],
        callback=MainWindow.paste,
    ),
    types.Action(
        id="about",
        icon=f"file:///{ABOUT_ICON_PATH}",
        title="About",
        status_tip="Show the application's About box",
        menus=[{"id": MenuId.HELP}],
        callback=MainWindow.about,
    ),
]


# Main setup

if __name__ == "__main__":
    app = Application(name="my_app")
    for action in ACTIONS:
        app.register_action(action)
    qapp = QApplication.instance() or QApplication([])
    qapp.setAttribute(Qt.ApplicationAttribute.AA_DontShowIconsInMenus)
    main_win = MainWindow(app=app)

    app.injection_store.register_provider(lambda: main_win, MainWindow)
    main_win.show()
    qapp.exec()
